import math
from math import pi, tanh
import numpy


class RasciiError(Exception):
    pass


class NonConvergenceError(RasciiError):
    pass


def check_breaking_criteria(height, depth, length):
    """
    Return two empty strings if everything is OK, else a string with
    warnings about breaking criteria and a string with warnings about
    being close to a breaking criterion
    """
    h1 = 0.14 * length
    h2 = 0.78 * depth
    h3 = 0.142 * tanh(2 * pi * depth / length) * length
    
    err = warn = ''
    for name, hmax in (('Length criterion', h1),
                       ('Depth criterion', h2),
                       ('Combined criterion', h3)):
        if height > hmax:
            err += '%s is exceeded, %.2f > %.2f\n' % (name, height, hmax)
        elif height > hmax * 0.9:
            warn += '%s is close to exceeded, %.2f = %.2f * %.3f\n' % \
                (name, height, hmax, height / hmax)
    return err, warn


def sinh_by_cosh(a, b):
    """
    A version of sinh(a)/cosh(b) where "b = a * f" and f is close
    to 1. This can then be written exp(a * (1 - f)) for large a
    """
    ans = numpy.zeros(a.size, float)
    for i, (ai, bi) in enumerate(zip(a, b)):
        if ai == 0:
            continue
        f = bi / ai
        if ((ai > 30 and 0.5 < f < 1.5) or (ai > 200 and 0.1 < f < 1.9)):
            ans[i] = math.exp(ai * (1 - f))
        else:
            sa = math.sinh(ai)
            cb = math.cosh(bi)
            ans[i] = sa / cb
    return ans


def cosh_by_cosh(a, b):
    """
    A version of cosh(a)/cosh(b) where "b = a * f" and f is close
    to 1. This can then be written exp(a * (1 - f)) for large a
    """
    ans = numpy.ones(a.size, float)
    for i, (ai, bi) in enumerate(zip(a, b)):
        if ai == 0:
            continue
        f = bi / ai
        if ((ai > 30 and 0.5 < f < 1.5) or (ai > 200 and 0.1 < f < 1.9)):
            ans[i] = math.exp(ai * (1 - f))
        else:
            ca = math.cosh(ai)
            cb = math.cosh(bi)
            ans[i] = ca / cb
    return ans


def blend_air_and_wave_velocities(x, z, t, wave, air, vel, eta_eps):
    """
    Compute velocities in the air phase and blend the water and air velocities
    in a divergence free manner from the free surface and a distance
    ``air.blending_height`` up. If this is ``None`` then blend all the way up to
    ``air.height``.
    
    The blending is done as follows. Introduce a new coordinate Z which is zero
    on the free surface and 1 at air_blend_distance above the free surface. Then
    the blending function is ``f = 3Z² - 2Z³`` which is used on the stream
    functions in the water and in the air. After taking the derivatives of this
    blended stream function the velocities are as can be seen in the code below.
    """
    eta = wave.surface_elevation(x, t)
    above = z > eta + eta_eps
    if air is not None and above.any():
        xa = x[above]
        za = z[above]
        ea = eta[above]
        vel_air = air.velocity(xa, za, t)
        
        blend = za < ea + air.blending_height
        if air.blending_height > 0 and blend.any():
            xb = xa[blend]
            zb = za[blend]
            eb = ea[blend]
            Z = (zb - eb) / air.blending_height
            vel_water = vel[above]
            psi_wave = wave.stream_function(xb, zb, t, frame='c')
            psi_air = air.stream_function(xb, zb, t, frame='c')
            detadx = wave.surface_slope(xb, t)
            
            if False:
                # Cubic smoothstep
                f = Z * Z * (3 - 2 * Z)
                dfdZ = 6 * Z - 6 * Z * Z
            else:
                # Fift order smootherstep
                f = Z * Z * Z * (Z * (Z * 6 - 15) + 10)
                dfdZ = Z * Z * (30 + Z * (30 * Z - 60))
            
            dZdx = -1 / air.blending_height * detadx
            dZdz = 1 / air.blending_height
            dfdx = dfdZ * dZdx
            dfdz = dfdZ * dZdz
            
            vel_air[blend, 0] = (1 - f) * vel_water[blend, 0] + f * vel_air[blend, 0]
            vel_air[blend, 1] = (1 - f) * vel_water[blend, 1] + f * vel_air[blend, 1]
            vel_air[blend, 0] += - dfdz * psi_wave + dfdz * psi_air
            vel_air[blend, 1] -= - dfdx * psi_wave + dfdx * psi_air
        vel[above] = vel_air
    else:
        vel[above] = 0


def blend_air_and_wave_velocity_cpp(wave_cpp, air_cpp, elevation_cpp, direction,
                                    eta_eps, air=None, psi_wave_cpp=None,
                                    psi_air_cpp=None, slope_cpp=None):
    """
    Return C++ code which blends the velocities in the water into the velocities
    in the air in such a way that the C++ code replicates the Python results
    from the blend_air_and_wave_velocities() function
    """
    if air is None:
        return 'x[2] < (%s) + %r ? (%s) : (%s)' % (elevation_cpp, eta_eps,
                                                   wave_cpp, '0.0')
    
    cpp = """[&]() {{
        const double elev = ({ecpp});
        
        const double val_water = ({uw_cpp});
        if (x[2] < elev + {eps!r}) {{
            // The point is below the free surface
            return val_water;
        }}
        
        const double dist_blend = {dist_blend!r};
        const double val_air = ({ua_cpp});
        if (x[2] < elev + dist_blend) {{
            // The point is in the blending zone
            const double Z = (x[2] - elev) / dist_blend;
            
            // Cubic smoothstep
            //const double f = Z * Z * (3 - 2 * Z);
            //const double dfdZ = Z * (6 - 6 * Z);
            
            // Fift order smootherstep
            const double f = Z * Z * Z * (Z * (Z * 6 - 15) + 10);
            const double dfdZ = Z * Z * (30 + Z * (30 * Z - 60));
            
            // Derivatives needed in the product rules
            const double dZdx = - 1.0 / dist_blend * ({slope_cpp});
            const double dZdz = 1.0 / dist_blend;
            const double dfdx = dfdZ * dZdx;
            const double dfdz = dfdZ * dZdz;
            
            return (1 - f) * val_water + f * val_air + %r * (%s);
        }}
        
        // The point is above the blending zone
        return val_air;
    }}()""".format(ecpp=elevation_cpp, uw_cpp=wave_cpp, ua_cpp=air_cpp,
                   eps=eta_eps, dist_blend=air.blending_height,
                   slope_cpp=slope_cpp)
    
    # Compute the product rule code
    if direction == 'x':
        sign = -1
        prcode = 'dfdz * (%s) - dfdz * (%s)' % (psi_wave_cpp, psi_air_cpp)
    elif direction == 'z':
        sign = +1
        prcode = 'dfdx * (%s) - dfdx * (%s)' % (psi_wave_cpp, psi_air_cpp)
    
    return cpp % (sign, prcode)



from numpy import pi, cos, sin, zeros, array, asarray, sinh, cosh, tanh
from .common import (blend_air_and_wave_velocities,
                     blend_air_and_wave_velocity_cpp)


class AiryWave:
    required_input = ('height', 'depth', 'length')
    optional_input = {'air': None, 'g': 9.81}
    
    def __init__(self, height, depth, length, air=None, g=9.81):
        """
        Linear Airy waves
        
        * height: wave height above still water level
        * depth: still water distance from the flat sea bottom to the surface
        * length: the periodic length of the wave (distance between peaks)
        """
        self.height = height
        self.depth = depth
        self.length = length
        self.air = air
        self.g = g
        self.order = 1
        self.warnings = ''
        
        self.k = 2 * pi / length
        self.omega = (self.k * g * tanh(self.k * depth))**0.5
        self.c = self.omega / self.k
        
        # For evaluating velocities close to the free surface
        self.eta_eps = self.height / 1e5
        
        # Provide velocities also in the air phase
        if self.air is not None:
            self.air.set_wave(self)
    
    def surface_elevation(self, x, t=0):
        """
        Compute the surface elavation at time t for position(s) x
        """
        if isinstance(x, (float, int)):
            x = array([x], float)
        x = asarray(x)
        return self.depth + self.height / 2 * cos(self.k * x - self.omega * t)
    
    def velocity(self, x, z, t=0, all_points_wet=False):
        """
        Compute the fluid velocity at time t for position(s) (x, z)
        where z is 0 at the bottom and equal to depth at the free surface
        """
        if isinstance(x, (float, int)):
            x, z = [x], [z]
        x = asarray(x, dtype=float)
        z = asarray(z, dtype=float)
        
        H = self.height
        k = self.k
        d = self.depth
        w = self.omega
        
        vel = zeros((x.size, 2), float) + 1
        vel[:, 0] = w * H / 2 * cosh(k * z) / sinh(k * d) * cos(k * x - w * t)
        vel[:, 1] = w * H / 2 * sinh(k * z) / sinh(k * d) * sin(k * x - w * t)
        
        if not all_points_wet:
            blend_air_and_wave_velocities(x, z, t, self, self.air, vel,
                                          self.eta_eps)
        
        return vel
    
    def elevation_cpp(self):
        """
        Return C++ code for evaluating the elevation of this specific wave.
        The positive traveling direction is x[0]
        """
        return '%r + %r / 2.0 * cos(%r * (x[0] - %r * t))' % \
            (self.depth, self.height, self.k, self.c)
    
    def velocity_cpp(self, all_points_wet=False):
        """
        Return C++ code for evaluating the particle velocities of this specific
        wave. Returns the x and z components only with z positive upwards. The
        positive traveling direction is x[0] and the vertical coordinate is x[2]
        which is zero at the bottom and equal to +depth at the mean water level.
        """
        H = self.height
        k = self.k
        d = self.depth
        w = self.omega
        
        cpp_x = '%r * cosh(%r * x[2]) * cos(%r * x[0] - %r * t)' %\
                (w * H / (2 * sinh(k * d)), k, k, w)
        cpp_z = '%r * sinh(%r * x[2]) * sin(%r * x[0] - %r * t)' %\
                (w * H / (2 * sinh(k * d)), k, k, w)
        
        if all_points_wet:
            return cpp_x, cpp_z
        
        # Handle velocities above the free surface
        e_cpp = self.elevation_cpp()
        cpp_ax = cpp_az = None
        cpp_psiw = cpp_psia = cpp_slope = None
        if self.air is not None:
            cpp_ax, cpp_az = self.air.velocity_cpp()
            cpp_psiw = self.stream_function_cpp(frame='c')
            cpp_psia = self.air.stream_function_cpp(frame='c')
            cpp_slope = self.slope_cpp()
            
        cpp_x = blend_air_and_wave_velocity_cpp(cpp_x, cpp_ax, e_cpp, 'x',
                                                self.eta_eps, self.air,
                                                cpp_psiw, cpp_psia, cpp_slope)
        cpp_z = blend_air_and_wave_velocity_cpp(cpp_z, cpp_az, e_cpp, 'z',
                                                self.eta_eps, self.air,
                                                cpp_psiw, cpp_psia, cpp_slope)
        
        return cpp_x, cpp_z



import math
from numpy import (pi, cos, sin, zeros, arange, trapz, isfinite, newaxis, array,
                   asarray, linspace, cosh, sinh)
from numpy.linalg import solve
from .common import (NonConvergenceError, sinh_by_cosh, cosh_by_cosh,
                     blend_air_and_wave_velocities,
                     blend_air_and_wave_velocity_cpp)


class FentonWave:
    required_input = {'height', 'depth', 'length', 'N'}
    optional_input = {'air': None, 'g': 9.81, 'relax': 0.5}
    
    def __init__(self, height, depth, length, N, air=None, g=9.81, relax=0.5):
        """
        Implement stream function waves based on the paper by Rienecker and
        Fenton (1981)
        
        * height: wave height above still water level
        * depth: still water distance from the flat sea bottom to the surface
        * length: the periodic length of the wave (distance between peaks)
        * N: the number of coefficients in the truncated Fourier series
        """
        self.height = height
        self.depth = depth
        self.length = length
        self.order = N
        self.air = air
        self.g = g
        self.relax = relax
        self.warnings = ''
        
        # Find the coeffients through optimization
        data = fenton_coefficients(height, depth, length, N, g, relax=relax)
        self.set_data(data)
        
        # For evaluating velocities close to the free surface
        self.eta_eps = self.height / 1e5
        
        # Provide velocities also in the air phase
        if self.air is not None:
            self.air.set_wave(self)
    
    def set_data(self, data):
        self.data = data
        self.eta = data['eta']         # Wave elevation at colocation points
        self.x = data['x']             # Positions of colocation points
        self.k = data['k']             # Wave number
        self.c = data['c']             # Phase speed
        self.cs = self.c - data['Q']   # Mean Stokes drift speed
        self.T = self.length / self.c  # Wave period
        
        # Cosine series coefficients for the elevation
        N = len(self.eta) - 1
        self.E = zeros(N + 1, float)
        J = arange(0, N + 1)
        self.E = trapz(self.eta * cos(J * J[:, newaxis] * pi / N))
    
    def stream_function(self, x, z, t=0, frame='b'):
        """
        Compute the stream function at time t for position(s) x
        """
        if isinstance(x, (float, int)):
            x, z = [x], [z]
        x2 = asarray(x, dtype=float) - self.c * t
        z2 = asarray(z, dtype=float)
        x2, z2 = x2[:, newaxis], z2[:, newaxis]
        
        N = len(self.eta) - 1
        B = self.data['B']
        k = self.k
        J = arange(1, N + 1)
        
        psi = (sinh(J * k * z2) / cosh(J * k * self.depth) *
               cos(J * k * x2)).dot(B[1:])
        
        if frame == 'b':
            return B[0] * z + psi
        elif frame == 'c':
            return psi
    
    def surface_elevation(self, x, t=0):
        """
        Compute the surface elevation at time t for position(s) x
        """
        if isinstance(x, (float, int)):
            x = array([x], float)
        x = asarray(x)
        
        # Cosine transformation of the elevation
        N = len(self.eta) - 1
        J = arange(0, N + 1)
        k, c = self.k, self.c
        return 2 * trapz(self.E * cos(J * k * (x[:, newaxis] - c * t))) / N
    
    def surface_slope(self, x, t=0):
        """
        Compute the x derivative of the surface elevation at time t
        """
        if isinstance(x, (float, int)):
            x = array([x], float)
        x = asarray(x)
        
        # Cosine transformation of the elevation
        N = len(self.eta) - 1
        J = arange(0, N + 1)
        k, c = self.k, self.c
        return - 2 * trapz(self.E * J * k *
                           sin(J * k * (x[:, newaxis] - c * t))) / N
    
    def velocity(self, x, z, t=0, all_points_wet=False):
        """
        Compute the fluid velocity at time t for position(s) (x, z)
        where z is 0 at the bottom and equal to depth at the free surface
        """
        if isinstance(x, (float, int)):
            x, z = [x], [z]
        x = asarray(x, dtype=float)
        z = asarray(z, dtype=float)
        
        N = len(self.eta) - 1
        B = self.data['B']
        k = self.k
        c = self.c
        J = arange(1, N + 1)
        
        vel = zeros((x.size, 2), float)
        vel[:, 0] = k * (B[1:] * cos(J * k * (x[:, newaxis] - c * t)) *
                         cosh(J * k * z[:, newaxis]) /
                         cosh(J * k * self.depth)).dot(J)
        vel[:, 1] = k * (B[1:] * sin(J * k * (x[:, newaxis] - c * t)) *
                         sinh(J * k * z[:, newaxis]) /
                         cosh(J * k * self.depth)).dot(J)
        
        if not all_points_wet:
            blend_air_and_wave_velocities(x, z, t, self, self.air, vel,
                                          self.eta_eps)
        
        return vel

    def stream_function_cpp(self, frame='b'):
        """
        Return C++ code for evaluating the stream function of this specific
        wave. The positive traveling direction is x[0] and the vertical
        coordinate is x[2] which is zero at the bottom and equal to +depth at
        the mean water level.
        """
        N = len(self.eta) - 1
        J = arange(1, N + 1)
        B = self.data['B']
        k = self.k
        c = self.c
        
        Jk = J * k
        facs = B[1:] / cosh(Jk * self.depth)
        
        cpp = ' + '.join('%r * cos(%f * (x[0] - %r * t)) * sinh(%r * x[2])' %
                         (facs[i], Jk[i], c, Jk[i]) for i in range(N))
        
        if frame == 'b':
            return '%r * x[2] + %s' % (B[0], cpp)
        elif frame == 'c':
            return cpp
    
    def elevation_cpp(self):
        """
        Return C++ code for evaluating the elevation of this specific wave.
        The positive traveling direction is x[0]
        """
        N = self.E.size - 1
        facs = self.E * 2 / N
        facs[0] *= 0.5
        facs[-1] *= 0.5
        code = ' + '.join('%r * cos(%d * %r * (x[0] - %r * t))' %
                          (facs[j], j, self.k, self.c)
                          for j in range(0, N + 1))
        return code
    
    def slope_cpp(self):
        """
        Return C++ code for evaluating the surface slope of this specific wave.
        The positive traveling direction is x[0]
        """
        N = self.E.size - 1
        facs = self.E * 2 / N * self.k * -1.0
        facs[0] *= 0.5
        facs[-1] *= 0.5
        code = ' + '.join('%r * %d * sin(%d * %r * (x[0] - %r * t))' %
                          (facs[j], j, j, self.k, self.c)
                          for j in range(0, N + 1))
        return code
    
    def velocity_cpp(self, all_points_wet=False):
        """
        Return C++ code for evaluating the particle velocities of this specific
        wave. Returns the x and z components only with z positive upwards. The
        positive traveling direction is x[0] and the vertical coordinate is x[2]
        which is zero at the bottom and equal to +depth at the mean water level.
        """
        N = len(self.eta) - 1
        J = arange(1, N + 1)
        B = self.data['B']
        k = self.k
        c = self.c
        
        Jk = J * k
        facs = J * B[1:] * k / cosh(Jk * self.depth)
        
        cpp_x = ' + '.join('%r * cos(%f * (x[0] - %r * t)) * cosh(%r * x[2])' %
                           (facs[i], Jk[i], c, Jk[i]) for i in range(N))
        cpp_z = ' + '.join('%r * sin(%f * (x[0] - %r * t)) * sinh(%r * x[2])' %
                           (facs[i], Jk[i], c, Jk[i]) for i in range(N))
        
        if all_points_wet:
            return cpp_x, cpp_z
        
        # Handle velocities above the free surface
        e_cpp = self.elevation_cpp()
        cpp_ax = cpp_az = None
        cpp_psiw = cpp_psia = cpp_slope = None
        if self.air is not None:
            cpp_ax, cpp_az = self.air.velocity_cpp()
            cpp_psiw = self.stream_function_cpp(frame='c')
            cpp_psia = self.air.stream_function_cpp(frame='c')
            cpp_slope = self.slope_cpp()
            
        cpp_x = blend_air_and_wave_velocity_cpp(cpp_x, cpp_ax, e_cpp, 'x',
                                                self.eta_eps, self.air,
                                                cpp_psiw, cpp_psia, cpp_slope)
        cpp_z = blend_air_and_wave_velocity_cpp(cpp_z, cpp_az, e_cpp, 'z',
                                                self.eta_eps, self.air,
                                                cpp_psiw, cpp_psia, cpp_slope)
        
        return cpp_x, cpp_z


def fenton_coefficients(height, depth, length, N, g=9.8, maxiter=500,
                        tolerance=1e-8, relax=1.0, num_steps=None):
    """
    Find B, Q and R by Newton-Raphson following Rienecker and Fenton (1981)
    
    Using relaxation can help in some difficult cases, try a value less than 1
    to decrease convergence speed, but increase chances of converging.
    """
    # Non dimensionalised input
    H = height / depth
    lam = length / depth
    k = 2 * pi / lam
    c = (math.tanh(k) / k)**0.5
    D = 1
    N_unknowns = 2 * (N + 1) + 2
    
    # Input data arrays
    J = arange(1, N + 1)
    M = arange(0, N + 1)
    x = M * lam / (2 * N)
    
    def initial_guess(H):
        """
        Initial guesses for the unknowns (linear wave)
        """
        B = zeros(N + 1, float)
        B[0] = c
        B[1] = -H / (4 * c * k)
        eta = 1 + H / 2 * cos(k * x)
        Q = c
        R = 1 + 0.5 * c**2
        return B, Q, R, eta
    
    def optimize(B, Q, R, eta, H):
        """
        Find B, Q and R by Newton iterations starting from the given initial
        guesses. According to Rienecker and Fenton (1981) a linear theory
        initial guess should work unless H close to breaking, then an initial
        guess from the optimization routine run with a slightly lower H should
        be used instead.
        """
        # Insert initial guesses into coefficient vector
        coeffs = zeros(N_unknowns, float)
        coeffs[:N + 1] = B
        coeffs[N + 1:2 * N + 2] = eta
        coeffs[2 * N + 2] = Q
        coeffs[2 * N + 3] = R
        f = func(coeffs, H, k, D, J, M)
        
        for it in range(1, maxiter + 1):
            jac = fprime(coeffs, H, k, D, J, M)
            delta = solve(jac, -f)
            coeffs += delta * relax
            f = func(coeffs, H, k, D, J, M)
            
            # Check the progress
            error = abs(f).max()
            eta_max = coeffs[N + 1:2 * N + 2].max()
            eta_min = coeffs[N + 1:2 * N + 2].min()
            if eta_max > 2:
                raise NonConvergenceError('Optimization did not converge. Got '
                                          'max(eta)/depth = %r in iteration %d' % (eta_max, it))
            elif eta_min < 0.1:
                raise NonConvergenceError('Optimization did not converge. Got '
                                          'min(eta)/depth = %r in iteration %d' % (eta_min, it))
            elif not isfinite(error):
                raise NonConvergenceError('Optimization did not converge. Got '
                                          'error %r in iteration %d' % (error, it))
            elif error < tolerance:
                B = coeffs[:N + 1]
                eta = coeffs[N + 1:2 * N + 2]
                Q = coeffs[2 * N + 2]
                R = coeffs[2 * N + 3]
                return B, Q, R, eta, error, it
        raise NonConvergenceError('Optimization did not converge after %d '
                                  'iterations, error = %r' % (it, error))
    
    # Perform the optimization, optionally in steps gradually increasing H
    steps = wave_height_steps(num_steps, D, lam, H)
    B, Q, R, eta = initial_guess(steps[0])
    for Hi in steps:
        B, Q, R, eta, error, niter = optimize(B, Q, R, eta, Hi)
    
    # Scale back to physical space
    B[0] *= (g * depth)**0.5
    B[1:] *= (g * depth**3)**0.5
    return {'x': x * depth,
            'eta': eta * depth,
            'B': B,
            'Q': Q * (g * depth**3)**0.5,
            'R': R * g * depth,
            'k': k / depth,
            'c': B[0],
            'error': error,
            'niter': niter}


def wave_height_steps(num_steps, D, lam, H):
    """
    Compute the breaking height and use this to select how many steps take when
    gradually increasing the wave height to improve convergence on high waves
    """
    # Breaking height
    Hb = 0.142 * math.tanh(2 * pi * D / lam) * lam
    
    # Try with progressively higher waves to get better initial conditions
    if num_steps is not None:
        pass
    if H > 0.75 * Hb:
        num_steps = 10
    elif H > 0.65 * Hb:
        num_steps = 5
    else:
        num_steps = 3
    
    if num_steps == 1:
        return [H]
    else:
        return linspace(H / num_steps, H, num_steps)


def func(coeffs, H, k, D, J, M):
    "The function to minimize"
    N_unknowns = coeffs.size
    N = J.size
    
    B0 = coeffs[0]
    B = coeffs[1:N + 1]
    eta = coeffs[N + 1:2 * N + 2]
    Q = coeffs[2 * N + 2]
    R = coeffs[2 * N + 3]
    
    # The function to me minimized
    f = zeros(N_unknowns, float)
    
    # Loop over the N + 1 points along the half wave
    for m in M:
        S1 = sinh_by_cosh(J * k * eta[m], J * k * D)
        C1 = cosh_by_cosh(J * k * eta[m], J * k * D)
        S2 = sin(J * m * pi / N)
        C2 = cos(J * m * pi / N)
        
        # Velocity at the free surface
        # The sign of B0 is swapped from what is in the paper
        um = -B0 + k * J.dot(B * C1 * C2)
        vm = 0 + k * J.dot(B * S1 * S2)
        
        # Enforce a streamline along the free surface
        # The sign of B0 is swapped from what is in the paper
        f[m] = -B0 * eta[m] + B.dot(S1 * C2) + Q
        
        # Enforce the dynamic free surface boundary condition
        f[N + 1 + m] = (um**2 + vm**2) / 2 + eta[m] - R
        
    # Enforce mean(eta) = D
    f[-2] = trapz(eta) / N - 1
        
    # Enforce eta_0 - eta_N = H, the wave height criterion
    f[-1] = eta[0] - eta[-1] - H
    
    return f


def fprime_num(coeffs, H, k, D, J, M):
    "The Jacobian of the function to minimize (numerical version)"
    N_unknowns = coeffs.size
    dc = 1e-10
    jac = zeros((N_unknowns, N_unknowns), float)
    f0 = func(coeffs, H, k, D, J, M)
    for i in range(N_unknowns):
        incr = zeros(N_unknowns, float)
        incr[i] = dc
        f1 = func(coeffs + incr, H, k, D, J, M)
        jac[:, i] = (f1 - f0) / dc
    return jac


def fprime(coeffs, H, k, D, J, M):
    "The Jacobian of the function to minimize"
    N_unknowns = coeffs.size
    N = J.size
    
    jac = zeros((N_unknowns, N_unknowns), float)
    B0 = coeffs[0]
    B = coeffs[1:N + 1]
    eta = coeffs[N + 1:2 * N + 2]
    
    for m in range(N + 1):
        S1 = sinh_by_cosh(J * k * eta[m], J * k * D)
        C1 = cosh_by_cosh(J * k * eta[m], J * k * D)
        S2 = sin(J * m * pi / N)
        C2 = cos(J * m * pi / N)
        
        SC = S1 * C2
        SS = S1 * S2
        CC = C1 * C2
        CS = C1 * S2
        
        # Velocity at the free surface
        um = -B0 + k * J.dot(B * CC)
        vm = 0 + k * J.dot(B * SS)
        
        # Derivatives of the eq. for the streamline along the free surface
        jac[m, N + 1 + m] = um
        jac[0:N + 1, 0] = -eta
        jac[m, 1:N + 1] = SC
        jac[m, -2] = 1
        
        # Derivatives of the dynamic free surface boundary condition
        jac[N + 1 + m, N + 1 + m] = 1 + (um * k**2 * B.dot(J**2 * SC) +
                                         vm * k**2 * B.dot(J**2 * CS))
        jac[N + 1 + m, -1] = -1
        jac[N + 1 + m, 0] = -um
        jac[N + 1 + m, 1:N + 1] = k * um * J * CC + k * vm * J * SS
    
    # Derivative of mean(eta) = 1
    jac[-2, N + 1:2 * N + 2] = M * 0 + 1 / N
    jac[-2, N + 1] = 1 / (2 * N)
    jac[-2, 2 * N + 1] = 1 / (2 * N)
    
    # Derivative of the wave height criterion
    jac[-1, N + 1] = 1
    jac[-1, 2 * N + 1] = -1
    
    return jac



from numpy import asarray, zeros


class ConstantAirPhase:
    def __init__(self, height, blending_height=None):
        """
        Constant horizontal velocity equal to the phase speed
        """
        self.height = height
        self.blending_height = blending_height
    
    def set_wave(self, wave):
        """
        Connect this air phase with the wave in the water phase
        """
        self.c = wave.c
        
        if self.blending_height is None:
            self.blending_height = 2 * wave.height
    
    def stream_function(self, x, z, t=0, frame='b'):
        """
        Compute the stream function at time t for position(s) x
        """
        if isinstance(x, (float, int)):
            x, z = [x], [z]
        z = asarray(z, dtype=float)
        
        if frame == 'e':
            return self.c * z
        elif frame == 'c':
            return 0.0 * z
    
    def velocity(self, x, z, t=0):
        """
        Compute the air phase particle velocity at time t for position(s) (x, z)
        where z is 0 at the bottom and equal to depth_water at the free surface
        and equal to depth_water + depth air at the top free slip lid above the
        air phase
        """
        if isinstance(x, (float, int)):
            x, z = [x], [z]
        x = asarray(x, dtype=float)
        z = asarray(z, dtype=float)
        
        return zeros((x.size, 2), float)
    
    def stream_function_cpp(self, frame='b'):
        """
        Return C++ code for evaluating the stream function of this specific
        wave. The positive traveling direction is x[0] and the vertical
        coordinate is x[2] which is zero at the bottom and equal to +depth at
        the mean water level.
        """
        if frame == 'b':
            return '%r * x[2]' % self.c
        elif frame == 'c':
            return '0.0'
    
    def velocity_cpp(self):
        """
        Return C++ code for evaluating the particle velocities of this specific
        wave. Returns the x and z components only with z positive upwards. The
        positive traveling direction is x[0] and the vertical coordinate is x[2]
        which is zero at the bottom and equal to +depth at the mean water level.
        """
        cpp_x = '0.0'
        cpp_z = '0.0'
        return (cpp_x, cpp_z)
    
    def __repr__(self):
        return ('ConstantAirPhase(height={s.height}, blending_height='
                '{s.blending_height})').format(s=self)



"""
Raschii is a library for constructing non-linear regular waves

Raschii is written by Tormod Landet (c) 2018- and is named after Thysanoessa
raschii, the Arctic Krill.

SPDX-License-Identifier: Apache-2.0
"""
__version__ = '1.0.0'
from .common import check_breaking_criteria, RasciiError, NonConvergenceError
from .airy import AiryWave
from .fenton import FentonWave
from .stokes import StokesWave
from .air_phase_fenton import FentonAirPhase
from .air_phase_constant import ConstantAirPhase


# The available wave models
WAVE_MODELS = {
    'Airy': AiryWave,
    'Fenton': FentonWave,
    'Stokes': StokesWave
}

# Air phase models
AIR_MODELS = {'FentonAir': FentonAirPhase,
              'ConstantAir': ConstantAirPhase}


def get_wave_model(model_name, air_model_name=None):
    """
    Get a Raschii wave model by name
    """
    if '+' in model_name:
        model_name, air_model_name = model_name.split('+')
    
    if model_name not in WAVE_MODELS:
        raise RasciiError('Wave model %r is not supported, supported wave '
                          'models are %s' % (model_name,
                                             ', '.join(WAVE_MODELS.keys())))
    wave = WAVE_MODELS[model_name]
    
    if air_model_name is None:
        return wave, None
    
    if air_model_name not in AIR_MODELS:
        raise RasciiError('Air model %r is not supported, supported air phase '
                          'models are %s' % (air_model_name,
                                             ', '.join(AIR_MODELS.keys())))
    air = AIR_MODELS[air_model_name]
    
    return wave, air



from numpy import zeros, asarray, arange, sin, cos, sinh, cosh, newaxis
from numpy.linalg import solve
from .common import sinh_by_cosh


class FentonAirPhase:
    def __init__(self, height, blending_height=None):
        """
        Given a set of colocation points with precomputed surface elevations
        obtained from a wave model in the water phase, produce a stream function
        approximation of the velocities in the air phase.
        """
        self.height = height
        self.blending_height = blending_height
    
    def set_wave(self, wave):
        """
        Connect this air phase with the wave in the water phase
        """
        N = wave.order
        self.x = arange(N + 1) * wave.length / (2 * N)
        self.eta = wave.surface_elevation(self.x)
        self.c = wave.c
        self.k = wave.k
        self.g = wave.g
        BQ = air_velocity_coefficients(self.x, self.eta, self.c, self.k,
                                       wave.depth, self.height)
        self.B, self.Q = BQ
        self.depth_water = wave.depth
        
        if self.blending_height is None:
            self.blending_height = 2 * wave.height
    
    def stream_function(self, x, z, t=0, frame='b'):
        """
        Compute the stream function at time t for position(s) x
        """
        if isinstance(x, (float, int)):
            x, z = [x], [z]
        x2 = asarray(x, dtype=float) - self.c * t
        z2 = self.depth_water + self.height - asarray(z, dtype=float)
        x2, z2 = x2[:, newaxis], z2[:, newaxis]
        
        N = len(self.eta) - 1
        B0 = self.c
        B = self.B
        k = self.k
        J = arange(1, N + 1)
        
        psi = (sinh(J * k * z2) / cosh(J * k * self.height) *
               cos(J * k * x2)).dot(B)
        
        if frame == 'e':
            return B0 * z + psi
        elif frame == 'c':
            return psi
    
    def velocity(self, x, z, t=0):
        """
        Compute the air phase particle velocity at time t for position(s) (x, z)
        where z is 0 at the bottom and equal to depth_water at the free surface
        and equal to depth_water + depth air at the top free slip lid above the
        air phase
        """
        if isinstance(x, (float, int)):
            x, z = [x], [z]
        x = asarray(x, dtype=float)
        z = asarray(z, dtype=float)
        
        B = self.B
        k = self.k
        c = self.c
        top = self.depth_water + self.height
        J = arange(1, B.size + 1)
        D = self.height
        x2 = x[:, newaxis] - c * t
        z2 = top - z[:, newaxis]
        
        vel = zeros((x.size, 2), float)
        vel[:, 0] = (k * B * cos(J * k * x2) * cosh(J * k * z2) /
                     cosh(J * k * D)).dot(J) * -1
        vel[:, 1] = (k * B * sin(J * k * x2) * sinh(J * k * z2) /
                     cosh(J * k * D)).dot(J)
        
        return vel
    
    def stream_function_cpp(self, frame='b'):
        """
        Return C++ code for evaluating the stream function of this specific
        wave. The positive traveling direction is x[0] and the vertical
        coordinate is x[2] which is zero at the bottom and equal to +depth at
        the mean water level.
        """
        N = len(self.eta) - 1
        J = arange(1, N + 1)
        k = self.k
        c = self.c
        
        Jk = J * k
        facs = self.B / cosh(Jk * self.height)
        
        z2 = '(%r - x[2])' % (self.depth_water + self.height, )
        cpp = ' + '.join('%r * cos(%f * (x[0] - %r * t)) * sinh(%r * %s)' %
                         (facs[i], Jk[i], c, Jk[i], z2) for i in range(N))
        
        if frame == 'b':
            B0 = self.c
            return '%r * x[2] + %s' % (B0, cpp)
        elif frame == 'c':
            return cpp
    
    def velocity_cpp(self):
        """
        Return C++ code for evaluating the particle velocities of this specific
        wave. Returns the x and z components only with z positive upwards. The
        positive traveling direction is x[0] and the vertical coordinate is x[2]
        which is zero at the bottom and equal to +depth at the mean water level.
        """
        N = len(self.eta) - 1
        J = arange(1, N + 1)
        k = self.k
        c = self.c
        
        Jk = J * k
        facs = J * self.B * k / cosh(Jk * self.height)
        
        z2 = '(%r - x[2])' % (self.depth_water + self.height, )
        cpp_x = ' + '.join('%r * cos(%f * (x[0] - %r * t)) * cosh(%r * %s)' %
                           (facs[i], Jk[i], c, Jk[i], z2) for i in range(N))
        cpp_z = ' + '.join('%r * sin(%f * (x[0] - %r * t)) * sinh(%r * %s)' %
                           (facs[i], Jk[i], c, Jk[i], z2) for i in range(N))
        return (cpp_x, cpp_z)
    
    def __repr__(self):
        return ('FentonAirPhase(height={s.height}, blending_height='
                '{s.blending_height})').format(s=self)


def air_velocity_coefficients(x, eta, c, k, depth_water, height_air):
    """
    This uses the same method as in M.M.Rienecker and J. D. Fenton (1981), but
    since the surface elvation and phase speed is known the problem is now
    linear in the unknowns B1..BN and Q
    """
    Nm = len(eta)
    Nj = Nm - 1
    Neq = Nm
    Nuk = Nj + 1
    J = arange(1, Nj + 1)
    D = height_air
    z = depth_water + height_air - eta
    
    lhs = zeros((Neq, Nuk), float)
    rhs = zeros(Neq, float)
    for m in range(Nm):
        S1 = sinh_by_cosh(J * k * z[m], J * k * D)
        C2 = cos(J * k * x[m])
        
        # The free surface is a stream line (stream func = const Q)
        lhs[m, :Nj] = S1 * C2
        lhs[m, -1] = 1
        rhs[m] = - c * z[m]
    
    BQ = solve(lhs, rhs)
    B = BQ[:-1]
    Q = BQ[-1]
    
    return B, Q



from math import pi, sinh, tanh, exp, sqrt, pow
import numpy
from .common import blend_air_and_wave_velocities


class StokesWave:
    required_input = {'height', 'depth', 'length', 'N'}
    optional_input = {'air': None, 'g': 9.81}
    
    def __init__(self, height, depth, length, N, air=None, g=9.81):
        """
        Implement Stokes waves based on the paper by J. D. Fenton (1985),
        "A Fifth‐Order Stokes Theory for Steady Waves".
        
        * height: wave height above still water level
        * depth: still water distance from the flat sea bottom to the surface
        * length: the periodic length of the wave (distance between peaks)
        * N: the number of coefficients in the truncated Fourier series
        """
        self.height = height
        self.depth = depth
        self.length = length
        self.order = N
        self.air = air
        self.g = g
        self.warnings = ''
        
        if N < 1:
            self.warnings = 'Stokes order must be at least 1, using order 1'
            self.order = 1
        elif N > 5:
            self.warnings = 'Stokes order is maximum 5, using order 5'
            self.order = 4
        
        # Find the coeffients through optimization
        self.k = 2 * pi / length  # The wave number
        data = stokes_coefficients(self.k * depth, self.order)
        self.set_data(data)
        
        # For evaluating velocities close to the free surface
        self.eta_eps = self.height / 1e5
        
        # Provide velocities also in the air phase
        if self.air is not None:
            self.air.set_wave(self)
    
    def set_data(self, data):
        self.data = data
        k = 2 * pi / self.length
        eps = k * self.height / 2
        d = self.depth
        
        c = (data['C0'] +
             pow(eps, 2) * data['C2'] +
             pow(eps, 4) * data['C4']) * sqrt(self.g / k)
        Q = (c * d * sqrt(k**3 / self.g) +
             data['D2'] * eps**2 +
             data['D4'] * eps**4) * sqrt(self.g / k**3)
        
        self.c = c        # Phase speed
        self.cs = c - Q   # Mean Stokes drift speed (TODO: verify this!)
        self.T = self.length / self.c  # Wave period
    
    def surface_elevation(self, x, t=0):
        """
        Compute the surface elavation at time t for position(s) x
        """
        if isinstance(x, (float, int)):
            x = numpy.array([x], float)
        x = numpy.asarray(x)
        
        d = self.depth
        k = self.k
        eps = k * self.height / 2
        kd = k * d
        D = self.data
        cos = numpy.cos
        eta = (kd + eps * cos(k * x) + eps**2 * D['B22'] * cos(2 * k * x) +
               eps**3 * D['B31'] * (cos(k * x) - cos(3 * k * x)) +
               eps**4 * (D['B42'] * cos(2 * k * x) +
                         D['B44'] * cos(4 * k * x)) +
               eps**5 * (-(D['B53'] + D['B55']) * cos(k * x) +
                         D['B53'] * cos(3 * k * x) +
                         D['B55'] * cos(5 * k * x))) / k
        return eta
    
    def velocity(self, x, z, t=0, all_points_wet=False):
        """
        Compute the fluid velocity at time t for position(s) (x, z)
        where z is 0 at the bottom and equal to depth at the free surface
        """
        if isinstance(x, (float, int)):
            x, z = [x], [z]
        x = numpy.asarray(x, dtype=float)
        z = numpy.asarray(z, dtype=float)
        x2 = x - self.c * t
        
        def my_cosh_cos(i, j):
            n = 'A%d%d' % (i, j)
            if self.data[n] == 0.0:
                return 0.0
            else:
                return pow(eps, i) * self.data[n] * j * self.k * numpy.cosh(
                    j * self.k * z) * numpy.cos(j * self.k * x2)

        def my_sinh_sin(i, j):
            n = 'A%d%d' % (i, j)
            if self.data[n] == 0.0:
                return 0.0
            else:
                return pow(eps, i) * self.data[n] * j * self.k * numpy.sinh(
                    j * self.k * z) * numpy.sin(j * self.k * x2)
        
        eps = self.k * self.height / 2
        vel = numpy.zeros((x.size, 2), float)
        vel[:, 0] = my_cosh_cos(1, 1) + my_cosh_cos(2, 2) + my_cosh_cos(3, 1) +\
            my_cosh_cos(3, 3) + my_cosh_cos(4, 2) + my_cosh_cos(4, 4) +\
            my_cosh_cos(5, 1) + my_cosh_cos(5, 3) + my_cosh_cos(5, 5)
        vel[:, 1] = my_sinh_sin(1, 1) + my_sinh_sin(2, 2) + my_sinh_sin(3, 1) +\
            my_sinh_sin(3, 3) + my_sinh_sin(4, 2) + my_sinh_sin(4, 4) +\
            my_sinh_sin(5, 1) + my_sinh_sin(5, 3) + my_sinh_sin(5, 5)
        vel *= self.data['C0'] * sqrt(self.g / self.k**3)
        
        if not all_points_wet:
            blend_air_and_wave_velocities(x, z, t, self, self.air, vel,
                                          self.eta_eps)
        
        return vel
    
    def elevation_cpp(self):
        """
        Return C++ code for evaluating the elevation of this specific wave.
        The positive traveling direction is x[0]
        """
        raise NotImplementedError()
    
    def velocity_cpp(self):
        """
        Return C++ code for evaluating the particle velocities of this specific
        wave. Returns the x and z components only with z positive upwards. The
        positive traveling direction is x[0] and the vertical coordinate is x[2]
        which is zero at the bottom and equal to +depth at the mean water level.
        """
        raise NotImplementedError()


def sech(x):
    "Hyperbolic secant"
    return (2 * exp(x) / (exp(2 * x) + 1))


def csch(x):
    "Hyperbolic cosecant"
    return (2 * exp(x) / (exp(2 * x) - 1))


def cotanh(x):
    "Hyperbolic cotangent"
    return (1 + exp(-2 * x)) / (1 - exp(-2 * x))


def stokes_coefficients(kd, N):
    """
    Define the Stokes expansion coefficients based on "A Fifth‐Order Stokes
    Theory for Steady Waves" (J. D. Fenton, 1985)
    
    The code uses pow instead of ** to be compatible with Dart
    """
    # Limit depth to 25 wave lengths to avoid overflow
    if kd > 50 * pi:
        kd = 50 * pi
    
    S = sech(2 * kd)
    Sh = sinh(kd)
    Th = tanh(kd)
    CTh = cotanh(kd)

    # Parameters are zero if not used in linear theory
    data = {'A11': csch(kd),
            'A22': 0.0,
            'A31': 0.0,
            'A33': 0.0,
            'A42': 0.0,
            'A44': 0.0,
            'A51': 0.0,
            'A53': 0.0,
            'A55': 0.0,
            'B22': 0.0,
            'B31': 0.0,
            'B42': 0.0,
            'B44': 0.0,
            'B53': 0.0,
            'B55': 0.0,
            'C0': sqrt(Th),
            'C2': 0.0,
            'C4': 0.0,
            'D2': 0.0,
            'D4': 0.0,
            'E2': 0.0,
            'E4': 0.0}

    if N == 1:
        return data

    # Define additional constants needed for second order Stokes waves
    data['A22'] = 3 * pow(S, 2) / (2 * pow(1 - S, 2))
    data['B22'] = CTh * (1 + 2 * S) / (2 * (1 - S))
    data['C2'] = sqrt(Th) * (2 + 7 * pow(S, 2)) / (4 * pow(1 - S, 2))
    data['D2'] = - sqrt(CTh) / 2
    data['E2'] = Th * (2 + 2 * S + 5 * pow(S, 2)) / (4 * pow(1 - S, 2))

    if N == 2:
        return data

    # Define additional constants needed for third order Stokes waves
    data['A31'] = (-4 - 20 * S + 10 * pow(S, 2) - 13 * pow(S, 3)) / (8 * Sh * pow(1 - S, 3))
    data['A33'] = (-2 * pow(S, 2) + 11 * pow(S, 3)) / (8 * Sh * pow(1 - S, 3))
    data['B31'] = -3 * (1 + 3 * S + 3 * pow(S, 2) + 2 * pow(S, 3)) / (8 * pow(1 - S, 3))

    if N == 3:
        return data

    # Define additional constants needed for forth order Stokes waves
    data['A42'] = (12 * S - 14 * pow(S, 2) - 264 * pow(S, 3) - 45 * pow(S, 4) - 13 * pow(S, 5)) /\
                  (24 * pow(1 - S, 5))
    data['A44'] = (10 * pow(S, 3) - 174 * pow(S, 4) + 291 * pow(S, 5) + 278 * pow(S, 6)) /\
                  (48 * (3 + 2 * S) * pow(1 - S, 5))
    data['B42'] = CTh * (6 - 26 * S - 182 * pow(S, 2) - 204 * pow(S, 3) -
                         25 * pow(S, 4) + 26 * pow(S, 5)) / (6 * (3 + 2 * S) * pow(1 - S, 4))
    data['B44'] = CTh * (24 + 92 * S + 122 * pow(S, 2) + 66 * pow(S, 3) +
                         67 * pow(S, 4) + 34 * pow(S, 5)) / (24 * (3 + 2 * S) * pow(1 - S, 4))
    data['C4'] = sqrt(Th) * (4 + 32 * S - 116 * pow(S, 2) - 400 * pow(S, 3) - 71 * pow(S, 4) +
                             146 * pow(S, 5)) / (32 * pow(1 - S, 5))
    data['D4'] = sqrt(CTh) * (2 + 4 * S + pow(S, 2) + 2 * pow(S, 3)) / (8 * pow(1 - S, 3))
    data['E4'] = Th * (8 + 12 * S - 152 * pow(S, 2) - 308 * pow(S, 3) - 42 * pow(S, 4) +
                       77 * pow(S, 5)) / (32 * pow(1 - S, 5))

    if N == 4:
        return data

    # Define additional constants needed for fifth order Stokes waves
    data['A51'] = (-1184 + 32 * S + 13232 * pow(S, 2) + 21712 * pow(S, 3) + 20940 * pow(S, 4) +
                   12554 * pow(S, 5) - 500 * pow(S, 6) - 3341 * pow(S, 7) - 670 * pow(S, 8)) /\
                  (64 * Sh * (3 + 2 * S) * (4 + S) * pow(1 - S, 6))
    data['A53'] = (4 * S + 105 * pow(S, 2) + 198 * pow(S, 3) - 1376 * pow(S, 4) - 1302 * pow(S, 5) -
                   117 * pow(S, 6) + 58 * pow(S, 7)) / (32 * Sh * (3 + 2 * S) * pow(1 - S, 6))
    data['A55'] = (-6 * pow(S, 3) + 272 * pow(S, 4) - 1552 * pow(S, 5) + 852 * pow(S, 6) +
                   2029 * pow(S, 7) + 430 * pow(S, 8)) / (64 * Sh * (3 + 2 * S) * (4 + S) * pow(1 - S, 6))
    data['B53'] = 9 * (132 + 17 * S - 2216 * pow(S, 2) - 5897 * pow(S, 3) - 6292 * pow(S, 4) -
                       2687 * pow(S, 5) + 194 * pow(S, 6) + 467 * pow(S, 7) + 82 * pow(S, 8)) /\
                      (128 * (3 + 2 * S) * (4 + S) * pow(1 - S, 6))
    data['B55'] = 5 * (300 + 1579 * S + 3176 * pow(S, 2) + 2949 * pow(S, 3) + 1188 * pow(S, 4) +
                       675 * pow(S, 5) + 1326 * pow(S, 6) + 827 * pow(S, 7) + 130 * pow(S, 8)) /\
                      (384 * (3 + 2 * S) * (4 + S) * pow(1 - S, 6))
    
    return data



from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
from codecs import open
import os, sys


here = os.path.abspath(os.path.dirname(__file__))


# Get the long description from the README file
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()
    

# Get the version
for line in open(os.path.join(here, 'raschii', '__init__.py'), encoding='utf-8'):
    if line.startswith('__version__'):
        version = line.split('=')[1].strip()[1:-1]


# List packages we depend on
dependencies = ['numpy']


# Make the setup.py test command work
class PyTest(TestCommand):
    description = 'Run Raschii\'s tests with pytest'
    
    def initialize_options(self):
        TestCommand.initialize_options(self)
    
    def finalize_options(self):
        TestCommand.finalize_options(self)
    
    def run_tests(self):
        import pytest
        args = ['-v', '--durations=10']
        if self.verbose:
            args.append('-s')
        args.append(os.path.join(here, 'tests/'))
        errno = pytest.main(args)
        sys.exit(errno)


# Give setuptools/pip informattion about the Ocellaris package
setup(
    name='raschii',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=version,

    description='An implementation of stream function wave theory in Python',
    long_description=long_description,

    # The project's main homepage.
    url='https://bitbucket.org/trlandet/raschii',

    # Author details
    author='Tormod Landet',
    author_email='tormod@landet.net',

    # Choose your license
    license='Apache 2.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Physics',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3'
    ],

    # What does your project relate to?
    keywords='ocean waves stream function Fenton Stokes Airy',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed.
    install_requires=dependencies,

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    package_data={},
    zip_safe=True,

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [],
    },
    
    # Configure the "test" command
    tests_require=['pytest'],
    cmdclass = {'test': PyTest},
)
